/**
 * GridView 循环列表
 * 目前仅支持从上往下滚动，单列
 * updateEvents: [function (node, index)]
 */
const { ccclass, property } = cc._decorator;

@ccclass
export default class GridView extends cc.Component {

    @property
    spacing = 0;

    @property
    gridSize = new cc.Vec2(100, 100);

    @property(cc.Prefab)
    itemPrefab: cc.Prefab = null;

    @property([cc.Component.EventHandler])
    updateEvents: cc.Component.EventHandler[] = [];

    @property(cc.ScrollView)
    _scrollView: cc.ScrollView = null;

    @property(cc.Node)
    _view: cc.Node = null;

    @property
    _capacity = 0;

    @property
    _visibleCount = 0;

    @property
    _visibleItems = null;

    @property
    _tempArray = null;

    @property
    _itemPoll = null;

    @property
    _target = null;

    onLoad() {

    }

    start() {
        // this.init(10);
    }

    onDestroy() {
        if (this._itemPoll != null) {
            this._itemPoll.clear();
        }
    }

    /**
     * 初始化
     * @param {*} capacity 容量
     */
    init(capacity, target) {
        this._target = target;
        this._scrollView = this.getComponent(cc.ScrollView);
        this._view = this._scrollView.node.getChildByName("view");
        this._visibleItems = {};
        this._tempArray = new Array();
        this._visibleCount = 1 + Math.ceil((this._view.height + this.spacing) / (this.gridSize.y + this.spacing));
        this._itemPoll = new cc.NodePool();
        for (let i = 0; i < this._visibleCount; i++) {
            this._itemPoll.put(cc.instantiate(this.itemPrefab));
        }

        const scrollViewEventHandler = new cc.Component.EventHandler();
        scrollViewEventHandler.target = this.node;
        scrollViewEventHandler.component = "GridView";
        scrollViewEventHandler.handler = "onScroll";
        scrollViewEventHandler.customEventData = null;

        this._scrollView.scrollEvents.push(scrollViewEventHandler);

        this._capacity = capacity;
        this._scrollView.content.width = this.gridSize.x;
        this._scrollView.content.height = this._capacity * this.gridSize.y + (this._capacity - 1) * this.spacing;

        this.updateView();
    }

    /**
     * 重载列表，可改变容量
     * @param {*} capacity
     */
    reload(capacity: number) {
        this._capacity = capacity;
        this._scrollView.content.height = this._capacity * this.gridSize.y + (this._capacity - 1) * this.spacing;

        // 回收当前可见的节点
        for (const key of Object.keys(this._visibleItems)) {
            this._itemPoll.put(this._visibleItems[key]);
            delete (this._visibleItems[key]);
        }

        this.updateView();
    }

    updateView() {
        // 可见范围的起始
        const startPos = this._scrollView.content.y - this._view.height * 0.5;
        let startIndex = Math.floor((startPos + this.spacing) / (this.gridSize.y + this.spacing));
        startIndex = Math.max(0, startIndex);
        startIndex = Math.min(startIndex, this._capacity - this._visibleCount + 1);
        if (startIndex < 0) {
            startIndex = 0;
        }

        let endIndex = startIndex + this._visibleCount - 1;
        if (endIndex >= this._capacity) {
            endIndex = this._capacity - 1;
        }
        // var endIndex = Math.min(this._capacity - 1, startIndex + this._visibleCount - 1);

        // 哪些变为不可见
        for (const key of Object.keys(this._visibleItems)) {
            const index = Number(key);
            if (index < startIndex || index > endIndex) {
                this._tempArray.push(key);
            }
        }
        if (this._tempArray.length > 0) {
            for (let i = 0; i < this._tempArray.length; i++) {
                const key = this._tempArray[i];
                this._itemPoll.put(this._visibleItems[key]);
                delete (this._visibleItems[key]);
            }
            this._tempArray.splice(0, this._tempArray.length);
        }

        // 哪些变为可见
        for (let i = startIndex; i <= endIndex; i++) {
            const key = i.toString();
            if (this._visibleItems[key] == null) {
                const item = this._itemPoll.get();
                this._visibleItems[key] = item;
                item.parent = this._scrollView.content;
                item.position = this.getItemPos(i);

                this.updateItem(item, i);
            }
        }
    }

    updateItem(item: any, index: number) {
        for (let i = 0; i < this.updateEvents.length; i++) {
            this.updateEvents[i].emit([item, index, this._target]);
        }
    }

    updateVisibleItems() {
        for (const key of Object.keys(this._visibleItems)) {
            for (let i = 0; i < this.updateEvents.length; i++) {
                this.updateEvents[i].emit([this._visibleItems[key], Number(key)]);
            }
        }
    }

    onScroll(scrollView, eventType, customEventData) {
        if (this._capacity > 0) {
            this.updateView();
        }
    }

    getItem(index: number) {
        return this._visibleItems[index.toString()];
    }

    getItemPos(index: number) {
        const y = (index + 0.5) * this.gridSize.y + index * this.spacing;
        return new cc.Vec2(0, -y);
    }

}
