import CommonUtil from "../util/CommonUtil";

const { ccclass, property } = cc._decorator;

@ccclass
export default class SafeAreaView extends cc.Component {

    @property(cc.Node)
    areas: cc.Node[] = [];
    @property(cc.Node)
    maskAreas: cc.Node[] = [];

    @property
    matchWidth: number = 1920;
    @property
    matchHeight: number = 960;

    private _minRatio: number = 9 / 18;
    private _maxRatio: number = 10 / 16;
    private _scale: number = 1;
    private _lastWidth: number = 0;
    private _lastHeight: number = 0;

    getScale(): number {
        return this._scale;
    }

    // LIFE-CYCLE CALLBACKS:

    onLoad() {
        const winSize = this.getVisibleSize();
        this._lastWidth = winSize.width;
        this._lastHeight = winSize.height;
        cc.log("onload Screen resolution changes to: " + winSize.width + "x" + winSize.height);
        this.resetSafeArea();
    }

    update(dt) {
        const winSize = this.getVisibleSize();
        if (this._lastWidth != winSize.width || this._lastHeight != winSize.height) {
            cc.log("Screen resolution changes to: " + winSize.width + "x" + winSize.height);
            this._lastWidth = winSize.width;
            this._lastHeight = winSize.height;
            this.resetSafeArea();
        }
    }

    resetSafeArea() {
        if (this.areas == null || this.areas.length < 1) {
            return;
        }
        this._scale = 1;
        // 实际屏幕宽高比，像素单位比值
        const winSize = this.getVisibleSize();
        const screenRatio = winSize.width / winSize.height;

        // 按宽度适配的实际高度值，UI单位
        let height = this.matchWidth / screenRatio;
        // 因为整体缩小了，所以高度需要放大才能重新填满屏幕
        height /= this._scale;

        const minHeight = this.matchWidth / this._maxRatio;    // 实际最小高度值，UI单位
        const maxHeight = this.matchWidth / this._minRatio;    // 实际最大高度值，UI单位
        // Debug.LogFormat("MinHeight: {0}, MaxHeight: {1}, CurHeight: {2}", minHeight, maxHeight, height);
        if (height < minHeight) {
            // 屏幕太扁，需要增加高度到minHeight，这样上下显示区域就会超出屏幕，所以需要整体缩小
            this._scale *= height / minHeight;
            height = minHeight;
        } else if (height > maxHeight) {
            // 屏幕太窄，需要减小高度到maxHeight
            height = maxHeight;
        }

        // Debug.LogFormat("Final width: {0}, height: {1}, scale: {2}, ratio: {3}", matchWidth, height, scale, matchWidth / height);
        const areaSize = cc.v2(this.matchWidth, height);
        // let areaScale = new cc.Vec3(this._scale, this._scale, 1);
        // cc.log("Reset SafeArea to  w: %s, h: %s, s: %s", areaSize.x, areaSize.y, this._scale);
        for (const area of this.areas) {
            // area.anchorMin = new Vector2(0.5f, 0.5f);
            // area.anchorMax = new Vector2(0.5f, 0.5f);

            area.width = areaSize.x;
            area.height = areaSize.y;
            area.scale = this._scale;
            // area.anchoredPosition = center;
        }
        this.resetMaskArea(areaSize, this._scale);
    }

    resetMaskArea(areaSize: cc.Vec2, areaScale: number) {
        if (this.maskAreas == null || this.maskAreas.length < 1) {
            return;
        }
        const maskSize = this.getVisibleSize();
        areaSize.x *= areaScale;
        areaSize.y *= areaScale;
        for (const maskArea of this.maskAreas) {

            const width = (maskSize.width - areaSize.x) * 0.5;
            const leftMask = CommonUtil.findAChild(maskArea, "Left");
            const rightMask = CommonUtil.findAChild(maskArea, "Right");
            this.setWidth(leftMask, width);
            this.setWidth(rightMask, width);

            const height = (maskSize.height - areaSize.y) * 0.5;

            const topMask = CommonUtil.findAChild(maskArea, "Top");
            const bottomMask = CommonUtil.findAChild(maskArea, "Bottom");

            this.setHeight(topMask, height);
            this.setHeight(bottomMask, height);
        }
    }

    private setWidth(mask: cc.Node, width: number) {
        if (width < 0) {
            width = 0;
        }
        mask.width = width;
    }

    private setHeight(mask: cc.Node, height: number) {

        if (height < 0) {
            height = 0;
        }
        mask.height = height;
    }
    // update (dt) {}

    private getVisibleSize() {
        return cc.view.getVisibleSize();
    }
}
