
import ConfigManager from "../manager/ConfigManager";
import ResourcesLoadMgr from "../manager/ResourcesLoadMgr";
import ScheduleQueue from "../manager/ScheduleQueue";
import { GameGlobal } from "../model/GameGlobal";
import ConfirmDialogUI from "../ui/common/ConfirmDialogUI";
import UIManager from "../uiFrame/UIManager";
import { LocalSave } from "./LocalSave";

export default class CommonUtil {

    static GetLocalData(key: string, defaultValue: string): string {
        return LocalSave.getData(key, defaultValue);
    }

    static SetLocalData(key: string, data: string) {
        LocalSave.saveData(key, data);
    }

    static addNodeToScene(node) {
        cc.director.getScene().addChild(node);
    }

    /**
     * 创建预设到UI层
     * @param {*} path
     * @param {*} callback
     */
    static createPrefab(path: string, callback: (node: cc.Node) => void = null, target: any = null) {
        CommonUtil.createPrefabToParent(path, UIManager.getInstance()._layerNormal, callback, target);
    }

    /**
     * 创建预设到指定父节点
     * @param {*} path
     * @param {*} parentNode
     * @param {*} callback
     * @param {*} target
     */
    static createPrefabToParent(path: string, parentNode: cc.Node, callback: (node: cc.Node) => void = null, target: any = null, parentPath: string = ""): void {
        UIManager.getInstance().blockInput(true);
        CommonUtil.loadPrefab(path, function (err, prefab) {
            UIManager.getInstance().blockInput(false);
            if (err != null) {
                cc.error(err);
            }

            let newNode = cc.instantiate(prefab);
            if (parentNode != null) {
                let parent = CommonUtil.findAChild(parentNode, parentPath);
                if (parent == null) {
                    parent = parentNode;
                }
                parent.addChild(newNode);
            } else {
                CommonUtil.addNodeToScene(newNode);
            }
            if (callback != null) {
                callback.call(target, newNode);
            }
        });
    }

    /**
     * 创建一个提示窗
     */
    static showDialogUI(callback: (res: boolean) => void, createCallback: (view: ConfirmDialogUI) => void) {
        ConfirmDialogUI.Create(callback, createCallback);
    }

    /**
     * 快速简洁的设置按钮点击事件
     * @param {*} node 视图类实例，一般是this
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} handler 处理函数名
     * @param {*} customEventData 自定义数据
     */
    static quickSetButtonClick(view, path: any, handler: any, customEventData: any = null) {
        CommonUtil.setButtonClick(view.node, path, view.node, cc.js.getClassName(view), handler, customEventData);
    }
    /**
     * 为一个按钮添设置新的击事件，效果和编辑器上添加一样
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} target 事件处理代码组件所属的节点，一般是预设根节点Node
     * @param {*} component 这个是代码文件名
     * @param {*} handler 处理函数名
     * @param {*} customEventData 自定义数据
     */
    static setButtonClick(node: cc.Node, path: string, target: cc.Node, component: string, handler: any, customEventData: any): void {
        CommonUtil.clearButtonClick(node, path);
        CommonUtil.addButtonClick(node, path, target, component, handler, customEventData);
    }

    /**
     * 清空按钮点击事件
     * @param {*} node
     * @param {*} path
     */
    static clearButtonClick(node: any, path: any) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            let button = child.getComponent(cc.Button);
            button.clickEvents.length = 0;
        }
    }
    /**
     * 触发点击一个按钮
     */
    static clickButton(node: any, path: string, params: any): void {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            let button = child.getComponent(cc.Button);
            for (let i = 0; i < button.clickEvents.length; i++) {
                let event = button.clickEvents[i];
                event.emit(params);
            }
        }
    }

    /**
     * 为一个按钮添加一个点击事件，效果和编辑器上添加一样
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} target 事件处理代码组件所属的节点，一般是预设根节点Node
     * @param {*} component 这个是代码文件名
     * @param {*} handler 处理函数名
     * @param {*} customEventData 自定义数据
     */
    static addButtonClick(node: any, path: any, target: any, component: any, handler: any, customEventData: any) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            let button = child.getComponent(cc.Button);

            let clickEventHandler = new cc.Component.EventHandler();
            clickEventHandler.target = target;
            clickEventHandler.component = component;
            clickEventHandler.handler = handler;
            clickEventHandler.customEventData = customEventData;

            button.clickEvents.push(clickEventHandler);
        }
    }

    /**
     * 得到当前时间毫秒数
     */
    static getNowMs() {
        if (GameGlobal.serverTime == null) {
            return new Date().getTime();
        }
        let nowTicks = new Date().getTime() - GameGlobal.cVTimeOffset;
        return nowTicks;
    }

    /**
     * 得到当前时间毫秒数
     */
    static getNowMsFormatServerTime() {
        let cNow = CommonUtil.getNowMs();
        if (GameGlobal.serverTime == null) {
            return cNow;
        }
        let nowTicks = cNow - CommonUtil.getTimeOffset() + GameGlobal.serverTimeOffset;
        return nowTicks;
    }

    /**
     * 计算与服务器UTC时间差
     */
    static setCVTimeOffset() {
        if (GameGlobal.serverTime == null) {
            return 0;
        }
        let nowSTicks = new Date(GameGlobal.serverTime).getTime() - GameGlobal.serverTimeOffset;
        let nowCTicks = new Date().getTime() - CommonUtil.getTimeOffset();
        GameGlobal.cVTimeOffset = nowCTicks - nowSTicks;
    }

    /**
     * 得到当前时间秒数
     */
    static getNowSecond() {
        return this.getNowMs() / 1000;
    }

    /**
     * 当前时间加秒数  返回毫秒数
     */
    static addNowSecond(seconds) {
        return this.getNowMs() + seconds * 1000;
    }

    /**
     * 根据毫秒数(服务器时间)，判断是不是今天
     */
    static isToday(ms) {
        let time = ms;
        if (GameGlobal.serverTimeOffset != null) {
            time = ms - GameGlobal.serverTimeOffset + CommonUtil.getTimeOffset();
        }
        let date = new Date(time);
        let now = new Date(CommonUtil.getNowMs());
        return date.getFullYear() === now.getFullYear() && date.getMonth() === now.getMonth() && date.getDate() === now.getDate();
    }

    /**
     * 返回与UTC时间差  毫秒
     */
    static getTimeOffset() {
        let timeOffset = -new Date().getTimezoneOffset();
        return timeOffset * 60 * 1000;
    }

    /**
     * 根据毫秒数，得到有效期是否是永久
     * @param {*} ms
     */

    static isPermanent(ms: any) {
        // let date = new Date(ms);
        // cc.log(date.toString());
        if ((ms - CommonUtil.getNowMs()) >= 10 * 365 * 24 * 3600 * 1000) {
            return true;
        }
        return false;
    }
    /**
     * 根据毫秒数，得到时间长度，格式00:00:00
     * @param {*} ms
     */
    static getTimeCountStr(ms: any) {
        let millisecond = ms % 1000;
        let seconds = (ms - millisecond) / 1000;

        let second = seconds % 60;
        let minutes = (seconds - second) / 60;

        let minute = minutes % 60;
        let hours = (minutes - minute) / 60;

        return (hours < 10 ? "0" + hours : hours)
            + ":" + (minute < 10 ? "0" + minute : minute)
            + ":" + (second < 10 ? "0" + second : second);
    }

    /**
     * 根据毫秒数，得到时间串
     * 格式
     * “2019-8-19 10:59”
     * @param {*} ms
     */
    static getTimeFormatStr(ms: any, onlyDate = false): string {
        let time = ms;
        if (GameGlobal.serverTimeOffset != null) {
            time = time - GameGlobal.serverTimeOffset + CommonUtil.getTimeOffset();
        }
        let date = new Date(time);
        // cc.log(date.toDateString());
        let year = date.getFullYear();
        let month = date.getMonth() + 1;
        let day = date.getDate();
        let hour = date.getHours();
        let min = date.getMinutes();
        let dateStr = year + "-" + month + "-" + day;
        if (onlyDate) {
            return dateStr;
        }
        return year + "-" + month + "-" + day + " " + hour + ":" + min;
    }

    static findAChildByType<T extends cc.Component>(node: cc.Node, path: string, type: { prototype: T }): T {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            return child.getComponent(type);
        }
        return null;
    }

    /**
     * 设置node.active
     * @param {*} node
     * @param {*} path
     * @param {*} active
     */
    static SetActive(node: cc.Node, path: string, active: boolean): void {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            child.active = active;
        }
    }

    static isActiveSelf(node: cc.Node, path: string): boolean {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            return child.active;
        }
        return false;
    }

    static SetLocalScale(node: cc.Node, path: string, scale: number): void {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            child.setScale(scale);
        }
    }

    static setEulerAngles(node: cc.Node, path: string, angle: cc.Vec3) {
        let child: cc.Node = CommonUtil.findAChild(node, path);
        if (child != null) {
            child.eulerAngles = angle;
        }
    }

    static SetAngle(node: cc.Node, path: string, angle: number) {
        let child: cc.Node = CommonUtil.findAChild(node, path);
        if (child != null) {
            child.angle = angle;
        }
    }

    static SetImageFillAmount(node: cc.Node, path: string, fillAmount: number) {
        let child: cc.Sprite = CommonUtil.findAChildByType(node, path, cc.Sprite);
        if (child != null) {
            child.fillRange = fillAmount;
        }
    }

    static PlayAnimator(node: cc.Node, path: string, name: string) {
        let child: cc.Animation = CommonUtil.findAChildByType(node, path, cc.Animation);
        if (child != null) {
            child.play(name);
        }
    }

    // normalizedTime:百分比
    static PlayAnimatorFrom(node: cc.Node, path: string, normalizedTime: number) {
        let child: cc.Animation = CommonUtil.findAChildByType(node, path, cc.Animation);
        if (child != null) {
            let current = child.currentClip;
            child.play(current.name, normalizedTime * current.duration);
        }
    }

    /**
     *
     * @param node
     * @param path
     * @param duration seconds
     * @param scale
     * @param easing http://docs.cocos.com/creator/manual/zh/extension/api/editor-framework/share/easing.html
     * backOut, linear
     */
    static DOScale(node: cc.Node, path: string, duration: number, scale: number, easing: string): cc.Tween {
        let target = this.findAChild(node, path);
        if (target == null) {
            cc.warn("!!! No child with path: " + path);
            return;
        }

        return cc.tween(target).to(duration, { scale }, { easing }).start();
    }

    /**
     *
     * @param node
     * @param path
     * @param duration seconds
     * @param position 停止跳跃之后的坐标
     * @param height 跳跃高度
     * @param jumps 跳跃次数
     */
    static DOJump(node: cc.Node, path: string, duration: number, position: cc.Vec2, height: number, jumps: number, easingAction: any) {
        let target = this.findAChild(node, path);
        if (target == null) {
            cc.warn("!!! No child with path: " + path);
            return;
        }

        let action = cc.jumpTo(duration, position, height, jumps).easing(easingAction());
        return target.runAction(action);
    }

    /**
     *
     * @param node
     * @param path
     * @param duration seconds
     * @param opacity 255
     * @param easing http://docs.cocos.com/creator/manual/zh/extension/api/editor-framework/share/easing.html
     * backOut, linear
     */
    static DoFade(node: cc.Node, path: string, opacity: number, duration: number, easing: string = "linear"): cc.Tween {
        let target = CommonUtil.findAChild(node, path);
        if (target == null) {
            cc.warn("!!! No child with path: " + path);
            return;
        }

        return cc.tween(target).to(duration, { opacity }, { easing }).start();
    }

    /**
     *
     * @param node
     * @param path
     * @param duration seconds
     * @param fromOpacity 255
     * @param toOpacity 255
     * @param easing http://docs.cocos.com/creator/manual/zh/extension/api/editor-framework/share/easing.html
     * backOut, linear
     */
    static DoFadeFromTo(node: cc.Node, path: string, fromOpacity: number, toOpacity: number, duration: number, easing: string = "linear"): cc.Tween {
        let target = CommonUtil.findAChild(node, path);
        if (target == null) {
            cc.warn("!!! No child with path: " + path);
            return;
        }
        if (duration == 0) {
            target.opacity = toOpacity;
        } else {
            target.opacity = fromOpacity;
            return cc.tween(target).to(duration, { opacity: toOpacity }, { easing }).start();
        }
    }

    /**
     *
     * @param node
     * @param path
     * @param position Vec3
     * @param duration seconds
     * @param easing http://docs.cocos.com/creator/manual/zh/extension/api/editor-framework/share/easing.html
     * @param callback
     * backOut, linear
     */
    static DOMove(node: cc.Node, path: string, position: cc.Vec2, duration: number, easing: string, callback: () => void = null): cc.Tween {
        let target = this.findAChild(node, path);
        if (target == null) {
            cc.warn("!!! No child with path: " + path);
            return;
        }

        return cc.tween(target)
            .to(duration, { position }, { easing })
            .call((): void => { if (callback != null) { callback(); } })
            .start();
    }

    /**
     * 修改控件锚点对应位置
     * @param node
     * @param path
     * @param duration seconds
     * @param positionX new X
     * @param easing http://docs.cocos.com/creator/manual/zh/extension/api/editor-framework/share/easing.html
     *  "linear" | "fade" | "quadIn" | "quadOut" | "quadInOut" | "quadOutIn" |
     *  "cubicIn" | "cubicOut" | "cubicInOut" | "cubicOutIn" |
     *  "quartIn" | "quartOut" | "quartInOut" | "quartOutIn" |
     *  "quintIn" | "quintOut" | "quintInOut" | "quintOutIn" |
     *  "sineIn" | "sineOut" | "sineInOut" | "sineOutIn" |
     *  "expoIn" | "expoOut" | "expoInOut" | "expoOutIn" |
     *  "circIn" | "circOut" | "circInOut" | "circOutIn" |
     *  "elasticIn" | "elasticOut" | "elasticInOut" | "elasticOutIn" |
     *  "backIn" | "backOut" | "backInOut" | "backOutIn" |
     *  "bounceIn" | "bounceOut" | "bounceInOut" | "bounceOutIn"
     */
    static DOAnchorPosX(node: cc.Node, path: string, positionX: number, duration: number, easing: string): cc.Tween {
        let target = this.findAChild(node, path);
        if (target == null) {
            cc.warn("!!! No child with path: " + path);
            return;
        }

        let newPos = new cc.Vec3(positionX, target.y);
        return cc.tween(target).to(duration, { position: newPos }, { easing }).start();
    }

    /**
     * 修改控件锚点对应位置
     * @param node
     * @param path
     * @param duration seconds
     * @param positionY new Y
     * @param easing http://docs.cocos.com/creator/manual/zh/extension/api/editor-framework/share/easing.html
     * backOut, linear
     */
    static DOAnchorPosY(node: cc.Node, path: string, positionY: number, duration: number, easing: string): cc.Tween {
        let target = this.findAChild(node, path);
        if (target == null) {
            cc.warn("!!! No child with path: " + path);
            return;
        }

        let newPos = new cc.Vec3(target.x, positionY);
        return cc.tween(target).to(duration, { position: newPos }, { easing }).start();
    }

    static DOKill(node: cc.Node, path: string): void {
        let target = CommonUtil.findAChild(node, path);
        if (target == null) {
            cc.warn("!!!Cant not find node in path: ", path);
            return null;
        }
        cc.tween(target).stop();
    }

    /**
     * 设置button.interactable
     * @param {*} node
     * @param {*} path
     * @param {*} interactable
     */
    static SetBtnInteractable(node: cc.Node, path: string, interactable: boolean): void {
        let child = CommonUtil.findAChildByType(node, path, cc.Button);
        if (child != null) {
            child.interactable = interactable;
        }
    }

    static setAnimatorEnabled(node: cc.Node, path: string, enabled: boolean): void {
        let child = CommonUtil.findAChildByType(node, path, cc.Animation);
        if (child != null) {
            child.enabled = enabled;
        }
    }

    /**
     * 设置图片
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} sprite 图片
     * @param {*} autoSize 是否需要自动设置比例大小
     */
    static setImage(node: cc.Node, path: string, spriteFrame: cc.SpriteFrame, autoSize: boolean = true): void {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            child.getComponent(cc.Sprite).spriteFrame = spriteFrame;
            if (autoSize && spriteFrame != null) {
                CommonUtil.preserveAspect(node, path);
            }
        }
    }

    /**
     * 设置图片
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} spriteName 图片名称，从索引读取
     * @param {*} autoSize 是否需要自动设置比例大小
     */
    static SetImage(node: cc.Node, path: string, spriteName: string, autoSize: boolean = true): void {
        let targetNode = CommonUtil.findAChild(node, path);
        CommonUtil.loadSprite(spriteName, function (_err: Error, spriteFrame: cc.SpriteFrame) {
            CommonUtil.setImage(targetNode, "", spriteFrame, autoSize);
        });
    }

    /**
     * 设置图片
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} spriteName 图片名称，从索引读取
     * @param {*} autoSize 是否需要自动设置比例大小
     */
    static SetRewardImage(node: cc.Node, path: string, id: string, autoSize: boolean = true): void {
        let config = ConfigManager.GetConfig("Currency")[id];
        CommonUtil.SetImage(node, path, config.image, autoSize);
    }

    /**
     * 设置文本
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} text 文本
     */
    static setLabel(node: cc.Node, path: string, text: string | number): void {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            let str = "";
            if (typeof (text) == "number") {
                str = text.toString();
            } else {
                str = text;
            }
            child.getComponent(cc.Label).string = str;
        }
    }

    static SetText(node: cc.Node, path: string, text: string | number): void {
        CommonUtil.setLabel(node, path, text);
    }

    /**
     * 设置Rich文本
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} text 文本
     */
    static setRichLabel(node: cc.Node, path: string, text: string): void {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            child.getComponent(cc.RichText).string = text;
        }
    }

    /**
     * 设置颜色
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} color color
     */
    static setColor(node: any, path: any, color: cc.Color, alpha = 255) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            child.color = color;
            child.opacity = alpha;
        }
    }

    /**
     * 保留图片原始宽高比，缩放node大小
     * @param {*} node
     */
    static preserveAspect(node: any, path) {
        let child = CommonUtil.findAChild(node, path);
        if (child == null) {
            return;
        }

        let sprite = child.getComponent(cc.Sprite);
        let spriteFrame = sprite.spriteFrame;

        let originSize = spriteFrame.getOriginalSize();

        if (child.width / child.height < originSize.width / originSize.height) {
            child.height = child.width / (originSize.width / originSize.height);
        } else {
            child.width = child.height * (originSize.width / originSize.height);
        }
    }

    /**
     * 播放指定节点下所有特效
     * @param {*} node
     * @param {*} path
     */
    static playParticle(node: any, path: any) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            let children = child.getComponentsInChildren(cc.ParticleSystem);
            for (let i = 0; i < children.length; i++) {
                children[i].resetSystem();
            }
        }
    }

    static playAnimation(node, path) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            child.getComponent(cc.Animation).play();
        }
    }

    static stopAnimation(node, path) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            child.getComponent(cc.Animation).stop();
        }
    }

    static forceUpdateWidget(node, path) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            let widget = child.getComponent(cc.Widget);
            if (widget != null) {
                widget.updateAlignment();
            }
        }
    }

    /**
     * 为一个Toggle添加一个点击事件，效果和编辑器上添加一样
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} target 事件处理代码组件所属的节点，一般是预设根节点Node
     * @param {*} component 这个是代码文件名
     * @param {*} handler 处理函数名
     * @param {*} customEventData 自定义数据
     */
    static addToggleClick(node: any, path: any, target: any, component: any, handler: any, customEventData: any) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            let toggle = child.getComponent(cc.Toggle);

            let clickEventHandler = new cc.Component.EventHandler();
            clickEventHandler.target = target;
            clickEventHandler.component = component;
            clickEventHandler.handler = handler;
            clickEventHandler.customEventData = customEventData;

            toggle.checkEvents.push(clickEventHandler);
        }
    }

    /**
     * 为一个Toggle添设置新的击事件，效果和编辑器上添加一样
     * @param {*} node 节点
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} target 事件处理代码组件所属的节点，一般是预设根节点Node
     * @param {*} component 这个是代码文件名
     * @param {*} handler 处理函数名
     * @param {*} customEventData 自定义数据
     */
    static setToggleClick(node: any, path: any, target: any, component: any, handler: any, customEventData: any) {
        CommonUtil.clearToggleClick(node, path);
        CommonUtil.addToggleClick(node, path, target, component, handler, customEventData);
    }

    /**
     * 快速简洁的设置Toggle点击事件
     * @param {*} view 视图类实例，一般是this
     * @param {*} path 如果设置的是子节点，传子节点路径
     * @param {*} handler 处理函数名
     * @param {*} customEventData 自定义数据
     */
    static quickSetToggleClick(view: any, path: any, handler: any, customEventData: any) {
        this.setToggleClick(view.node, path, view.node, cc.js.getClassName(view), handler, customEventData);
    }

    /**
     * 清空Toggle点击事件
     * @param {*} node
     * @param {*} path
     */
    static clearToggleClick(node: any, path: any) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            let toggle = child.getComponent(cc.Toggle);
            toggle.checkEvents.length = 0;
        }
    }

    /**
     * 点击一个Toggle
     */
    static clickToggle(node, path, params = null) {
        let child = CommonUtil.findAChild(node, path);
        if (child != null) {
            let toggle = child.getComponent(cc.Toggle);
            toggle.isChecked = true;
            // for (let i = 0; i < toggle.checkEvents.length; i++) {
            //     let event = toggle.checkEvents[i];
            //     cc.log(event)
            //     event.emit(params);
            // }
        }
    }

    /**
     * 带屏蔽输入的scheduleOnce
     * @param {*} component
     * @param {*} func
     * @param {*} delay
     */
    static scheduleOnce(component: any, func: any, delay: any) {
        UIManager.getInstance().blockInput(true);
        component.scheduleOnce(function () {
            UIManager.getInstance().blockInput(false);
            func.call(component);
        }, delay);
    }

    /**
     * 添加一个延迟调用到队列中
     * @param {*} component
     * @param {*} func
     * @param {*} delay
     */
    static scheduleQueue(component: any, func: any, delay: any) {
        let queue = component.node.getComponent(ScheduleQueue);
        if (queue == null) {
            queue = component.node.addComponent(ScheduleQueue);
        }
        queue.add(component, func, delay);
    }

    /**
     * 播放一下音频
     * @param {*} name
     */
    static playSound(name: any) {
        cc.loader.loadRes("sounds/" + name, cc.AudioClip, function (err, clip) {
            if (clip != null) {
                cc.audioEngine.play(clip, false, 1);
            }
        });
    }

    static isAndroid() {
        return cc.sys.os == cc.sys.OS_ANDROID;
    }

    static isIOS() {
        return cc.sys.os == cc.sys.OS_IOS || cc.sys.os == cc.sys.OS_OSX;
    }

    static isWindows() {
        return cc.sys.os == cc.sys.OS_WINDOWS || cc.sys.os == cc.sys.OS_WINRT || cc.sys.os == cc.sys.OS_WP8;
    }

    static getOS() {
        if (this.isAndroid()) {
            return "ANDROID";
        } else if (this.isIOS()) {
            return "IOS";
        } else {
            return "OTHER";
        }
    }

    static findAChild(parent: cc.Node, childName: string): cc.Node {
        if (CommonUtil.stringIsEmpty(childName)) {
            return parent;
        }
        let child: cc.Node = cc.find(childName, parent);
        if (child == null) {
            for (const node of parent.children) {
                child = CommonUtil.findAChild(node, childName);
                if (child != null) {
                    return child;
                }
            }
        }
        return child;
    }

    static getChildAt(parent: cc.Node, path: string, index: number): cc.Node {
        let node = this.findAChild(parent, path);
        if (node == null || node.children.length - 1 < index) {
            return null;
        }
        return node.children[index];
    }

    static GetChildrenByType<T extends cc.Component>(node: cc.Node, path: string, type: { prototype: T }): T[] {
        let target = CommonUtil.findAChild(node, path);
        let ret: T[] = [];
        for (let child of target.children) {
            let v = child.getComponent(type);
            if (null != v) {
                ret.push(v);
            }
        }
        return ret;
    }

    static stringIsEmpty(str: string): boolean {
        return str == "" || str == null;
    }

    static getSignedNumString(num: number, needAddChar = true): string {
        let addChar = needAddChar ? "+" : "";
        if (num == null) {
            return "--";
        }
        if (num > 0) {
            return addChar + num;
        }
        return num.toString();
    }

    static removeAllChild(parent: cc.Node, ...exceptObjName) {
        for (const iterator of parent.children) {
            if (exceptObjName.find(function (v) { return v == iterator.name; }) == null) {
                iterator.destroy();
            }
        }
    }

    static removeAllChildByName(parent: cc.Node, ...includeObjName) {
        for (const iterator of parent.children) {
            if (includeObjName.find(function (v) { return v == iterator.name; }) != null) {
                iterator.destroy();
            }
        }
    }

    static getUserName(name: string, fixed_length: number = null): string {
        if (fixed_length == null) {
            fixed_length = 12;
        }
        if (name == null || name == "") {
            return "";
        }

        let length = 0;
        let i_length = 0;
        let index = 0;

        for (let i = 0; i < name.length; i++) {
            let x = name.charCodeAt(i);
            length += 1.5;
            if (x > 255) {
                length += 0.5;
            }

            if (length + 2 <= fixed_length) {
                i_length = length;
                index = i;
            }
        }

        let ret = name;
        if (length > fixed_length) {
            ret = name.substring(0, index + 1) + "..";
            if (i_length + 2 < fixed_length) {
                ret += ".";
            }
            return ret;
        }
        return ret;
    }

    static loadSprite(name: string, callBack) {
        cc.loader.loadRes(ConfigManager.GetSpritePath(name), cc.SpriteFrame, callBack);
    }

    static loadSpriteArr(nameArr, callBack) {
        let pathList = [];
        for (const name of nameArr) {
            pathList.push(ConfigManager.GetSpritePath(name));
        }
        cc.loader.loadResArray(pathList, cc.SpriteFrame, callBack);
    }

    static loadPrefab(name, callBack: (_err: Error, _node: cc.Node) => void = null) {
        ResourcesLoadMgr.getInstance().getPrefab(name, callBack);
    }

    static loadPrefabArr(nameArr: string[], callBack) {
        ResourcesLoadMgr.getInstance().getPrefabList(nameArr, callBack);
    }

    static loadResArr(urls: string[], progressCallback: (completedCount: number, totalCount: number, item: any) => void, completeCallback: (error: Error[], _res: Map<string, any>) => void): void {
        let index = 0;
        let err: Error[] = [];
        let resArr: Map<string, any> = new Map<string, any>();
        if (urls.length == 0) {
            completeCallback(err, resArr);
            return;
        }
        let load = (_index: number): void => {
            cc.loader.loadRes(urls[_index], (error: Error, res: any) => {
                if (error !== null) {
                    err.push(error);
                }
                if (res != null) {
                    resArr.set(res.name, res);
                }
                if (_index < urls.length - 1) {
                    progressCallback(index + 1, urls.length, res);
                    load(++index);
                } else {
                    completeCallback(err, resArr);
                }
            });
        };
        load(index);
    }

    static getItemTimeString(hours: number): string {
        if (hours >= 24) {
            return Math.floor(hours / 24) + "天";
        }
        return hours + "小时";
    }

    static getShortenDigitNumString(num) {
        if (num < 10000) {
            return num;
        }
        if (num >= 10000 && num < 100000) {
            return Math.trunc((num / 10000) * 100) / 100 + "万";
        }
        if (num >= 100000 && num < 1000000) {
            return Math.trunc((num / 10000) * 10) / 10 + "万";
        }
        return Math.floor(num / 10000) + "万";
    }

    static getItemNum(num, type, addStr = "") {
        return addStr + CommonUtil.getShortenDigitNumString(num);
    }

    static instantiateNode(node, parent, resetPos = true) {
        let nodeClone = cc.instantiate(node);
        parent.addChild(nodeClone);
        nodeClone.name = node.name + "Clone";
        nodeClone.active = true;
        if (resetPos) {
            nodeClone.setPosition(cc.Vec2.ZERO);
        }
        return nodeClone;
    }

    static removeRedSpot(target: cc.Node) {
        let redSpot = CommonUtil.findAChild(target, "RedSpot");
        if (redSpot != null) {
            redSpot.active = false;
        }
    }

    static createAndSetRedSpot(target: cc.Node) {
        let redSpot = CommonUtil.findAChild(target, "RedSpot");
        if (redSpot != null) {
            redSpot.active = true;
            return;
        } else {
            CommonUtil.loadPrefab("RedSpot", function (err, prefab) {
                redSpot = cc.instantiate(prefab);
                redSpot.name = "RedSpot";
                let parent = CommonUtil.findAChild(target, "RedSpotNode");
                if (parent != null) {
                    parent.addChild(redSpot);
                    let widget = redSpot.getComponent(cc.Widget);
                    widget.top = 0;
                    widget.right = 0;
                } else {
                    parent = target;
                    parent.addChild(redSpot);
                    let widget = redSpot.getComponent(cc.Widget);
                    widget.top = 0;
                    widget.right = -redSpot.width / 2;
                }
            });
        }
    }

    /**
     * 清除原有事件，覆盖为新的
     * @param node
     * @param path
     * @param action
     * @param target action对象
     * @param args
     */
    static setBtnClick(node: cc.Node, path: string, action: (...args: any[]) => void, target = null, ...args) {
        let btn = CommonUtil.findAChildByType(node, path, cc.Button);
        if (btn == null) {
            cc.log("no button ", path);
            return;
        }
        btn.node.off("click");
        btn.node.on("click", function () {
            SafeCall(target, action, ...args);
        });
    }

    static emitBtnClick(node, path: string, action = null) {
        let btn = CommonUtil.findAChildByType(node, path, cc.Button);
        if (btn == null) {
            cc.log("no button ", path);
            return;
        }
        let clickEvents = btn.clickEvents;
        for (let index = 0; index < clickEvents.length; index++) {
            const element = clickEvents[index];
            element.emit(["click"]);
        }
    }

    static accuracyAdd(arg1: number, arg2: number) {
        let r1 = 0;
        let s1 = arg1.toString();
        let arg1Arr = s1.split(".");
        if (arg1Arr.length > 1) {
            r1 = arg1Arr[1].length;
        }
        let r2 = 0;
        let s2 = arg2.toString();
        let arg2Arr = s2.split(".");
        if (arg2Arr.length > 1) {
            r2 = arg2Arr[1].length;
        }
        let m = Math.max(r1, r2);
        for (let index = r1; index < m; index++) {
            s1 = s1 + "0";
        }
        for (let index = r2; index < m; index++) {
            s2 = s2 + "0";
        }
        return (Number(s1.replace(".", "")) + Number(s2.replace(".", ""))) / Math.pow(10, m);
    }

    static accuracyMul(arg1: number, arg2: number) {
        let m = 0;
        let s1 = arg1.toString();
        let arg1Arr = s1.split(".");
        if (arg1Arr.length > 1) {
            m += arg1Arr[1].length;
        }
        let s2 = arg2.toString();
        let arg2Arr = s2.split(".");
        if (arg2Arr.length > 1) {
            m += arg2Arr[1].length;
        }
        return Number(s1.replace(".", "")) * Number(s2.replace(".", "")) / Math.pow(10, m);
    }

    static isButton(node: cc.Node): boolean {
        return node.getComponent(cc.Button) != null;
    }

    /**
     * Random int between min and max
     * @param min
     * @param max
     */
    static random(min: number, max: number): number {
        return Math.floor(Math.random() * (max - min)) + min;
    }

    static GetSystemLanguage(): string {
        // TODO
        return "";
    }

    static GetTime(): number {
        return new Date().getTime() / 1000;
    }

    static GetPastTime(time: number): number {
        return CommonUtil.GetTime() - time;
    }

    static GetEarlyTime(time: number): number {
        return -this.GetPastTime(time);
    }
    static GetAnimationLength(node: cc.Node, path: string, index: number): number {
        let child = CommonUtil.findAChildByType(node, path, cc.Animation);
        if (child != null) {
            let clips = child.getClips();
            if (clips != null && clips.length > index) {
                return clips[index].duration;
            }
        }
        return 0;
    }

    static SetComponentEnabled<T extends cc.Component>(node: cc.Node, path: string, type: { prototype: T }, enabled: boolean) {
        let com = CommonUtil.findAChildByType(node, path, type);
        if (com != null) {
            com.enabled = enabled;
        }
    }

    static SetComponentEnabledByComponent(com: cc.Component, enabled: boolean) {
        if (com != null) {
            com.enabled = enabled;
        }
    }

    static SetParent(child: cc.Node, parent: cc.Node) {
        if (child != null && parent != null) {
            child.parent = parent;
        }
    }

    static GetUIHeight(parent: cc.Node, path: string) {
        let node = CommonUtil.findAChild(parent, path);
        return node != null ? node.height : 0;
    }

    static SetUIHeight(parent: cc.Node, path: string, height: number) {
        let node = CommonUtil.findAChild(parent, path);
        if (node != null) {
            node.y = height;
        }
    }

    static GetCount(obj: object): number {
        return Object.keys(obj).length;
    }

    static ForceLayout(node: cc.Node, path: string) {
        let com = CommonUtil.findAChildByType(node, path, cc.Layout);
        if (com != null) {
            com.updateLayout();
        }
    }

    static DoScrollRectPercentY(node: cc.Node, path: string, index: number, totalCount: number, duration: number, easing: string) {
        let scrollRect = CommonUtil.findAChildByType(node, path, cc.ScrollView);
        if (scrollRect != null) {
            let content = scrollRect.content;
            let viewport = CommonUtil.findAChild(scrollRect.node, "view");

            let minLimit = viewport.y / 2;
            let maxLimit = minLimit + content.y;
            let y = minLimit;
            if (totalCount > 0) {
                y = minLimit + (content.y / totalCount) * index;
            }
            if (y < minLimit) {
                y = minLimit;
            }
            if (y > maxLimit) {
                y = maxLimit;
            }
            if (duration == 0) {
                content.setPosition(cc.v2(content.x, y));
            } else {
                CommonUtil.DOAnchorPosY(content, "", y, duration, easing);
            }
        }
    }

    static SetAnchoredPositionY(parent: cc.Node, path: string, y: number) {
        let node = CommonUtil.findAChild(parent, path);
        node.y = y;
    }

    /**
     * @param node
     * @param dir 0: UP, 1: DOWN, 2: LEFT, 3: RIGHT
     */
    static UpdateZOrder(node: cc.Node, dir: number = 1): void {
        switch (dir) {
            case 0: // UP
                node.zIndex = node.y;
                break;
            case 1: // DOWN
                node.zIndex = -node.y;
                break;
            case 2: // LEFT
                node.zIndex = node.x;
                break;
            case 3: // RIGHT
                node.zIndex = -node.x;
                break;
            default:
                break;
        }
    }

    /**
     * 转换原始容器中的pos坐标到新容器中
     * @param srcContainer
     * @param targetContainer
     * @param pos
     */
    static ConvertPosAR(srcContainer: cc.Node, targetContainer: cc.Node, pos: cc.Vec2): cc.Vec2 {
        let wPos = srcContainer.convertToWorldSpaceAR(pos);
        let lPos = targetContainer.convertToNodeSpaceAR(wPos);
        return lPos;
    }
}
