/* global Editor*/

const fs = require("fs");
const fse = require('fs-extra');
const path = require("fire-path");
const buildTools = require("./buildTools");
const utils = require("./buildUtils");
const Const = require("./Const");

let self = {

    wxSrcDir: path.join(Editor.Project.path, "build", "wechatgame", "res"),
    wxDstDir: path.join(Editor.Project.path, "build", "WX_DST", "res"),

    load: function () {
        // execute when package loaded
        // 在构建结束 之前 触发，此时除了计算文件 MD5、生成 settings.js、原生平台的加密脚本以外，大部分构建操作都已执行完毕。我们通常会在这个事件中对已经构建好的文件做进一步处理。
        Editor.Builder.on('before-change-files', buildTools.onBeforeBuildFinish);
    },

    unload: function () {
        // execute when package unloaded
        Editor.Builder.removeListener('before-change-files', buildTools.onBeforeBuildFinish);
    },

    // register your ipc messages here
    messages: {
        'debugPanel'() {
            Editor.Panel.open('build_index');
        },
        'createSpriteIndex'() {

            var includeExt = [".png", ".jpeg", ".jpg"];
            var dirPath = path.join(Editor.Project.path, 'assets', 'resources', 'texture');
            var saveData = {};
            let files = fs.readdirSync(dirPath);
            for (const filename of files) {
                var extension = path.extname(filename);
                var basePath = path.join(dirPath, filename);
                if (extension == ".meta") {
                    continue;
                } else {
                    self.fileRead(basePath, includeExt, saveData, 'texture');
                }
            }
            Editor.log("CreateSpriteIndex: " + JSON.stringify(saveData));
            var url = 'db://assets/resources/sysConfig/SpriteIndex.json';
            if (Editor.assetdb.exists(url)) {
                Editor.assetdb.saveExists(url, JSON.stringify(saveData), function (err) {
                    if (err) {
                        Editor.error(err);
                    }
                    Editor.log("Write File: ", url);
                });
            } else {
                fse.ensureDirSync(path.dirname(url));
                Editor.assetdb.create(url, JSON.stringify(saveData), function (err) {
                    if (err) {
                        Editor.error(err);
                    }
                    Editor.log("Write File: ", url);
                });
            }

        },

        'createPrefabIndex'() {
            var includeExt = [".prefab"];
            var dirPath = path.join(Editor.Project.path, 'assets', 'resources', 'prefab');
            var saveData = {};
            let files = fs.readdirSync(dirPath);
            for (const filename of files) {
                var extension = path.extname(filename);
                var basePath = path.join(dirPath, filename);
                if (extension == ".meta") {
                    continue;
                } else {
                    self.fileRead(basePath, includeExt, saveData, 'prefab');
                }
            }
            Editor.log("CreatePrefabIndex: " + JSON.stringify(saveData));
            var url = 'db://assets/resources/sysConfig/PrefabIndex.json';
            if (Editor.assetdb.exists(url)) {
                Editor.assetdb.saveExists(url, JSON.stringify(saveData), function () {
                    Editor.log("Write File: ", url);
                });
            } else {
                fse.ensureDirSync(path.dirname(url));
                Editor.assetdb.create(url, JSON.stringify(saveData), function () {
                    Editor.log("Write File: ", url);
                });
            }

        },

        'processWXProject'(event) {
            let excludePathArr = [
                "loading"
            ];
            // Collect all UUID
            let uuidList = [];
            for (const ePath of excludePathArr) {
                let fPath = path.join(Editor.Project.path, "assets", ePath);
                utils.walkFiles(fPath, (f, oList) => {
                    if (utils.isMeta(f)) {
                        let uuid = utils.getUUID(f);
                        if (uuid == undefined) {
                            Editor.warn('No UUID in file :' + f);
                            return;
                        }
                        oList.push(uuid);
                        // Editor.log('Processing UUID :' + uuid);
                    }
                }, uuidList);
            }
            Editor.log('Processing End Size: ' + uuidList.length);
            self.moveWXFile(uuidList);

            self.uploadWXFile();
            if (event.reply) {
                //if no error, the first argument should be null
                event.reply(null, 'Processing End Size: ' + uuidList.length);
            }

        },

        'uploadWXProject'() {
            self.uploadWXFile();
        }
    },

    fileRead: function (dirPath, includeExt, saveData, type) {
        let files = fs.readdirSync(dirPath);
        for (const file of files) {
            var extension = path.extname(file);
            var filePath = path.join(dirPath, file);
            if (utils.isMeta(filePath)) {
                continue;
            } else if (extension == "") {
                this.fileRead(filePath, includeExt, saveData, type);
            } else {

                for (const iterator of includeExt) {
                    if (iterator == extension) {
                        var filename = filePath.substring(filePath.lastIndexOf('\\') + 1, filePath.lastIndexOf('.'));
                        var resourcesPath = filePath.substring(filePath.lastIndexOf('\\' + type), filePath.lastIndexOf('.'));
                        var index = 0;
                        while (resourcesPath.indexOf('\\') != -1 && index < 10) {
                            resourcesPath = resourcesPath.replace('\\', '/');
                            index++;
                        }
                        if (saveData[filename] != null) {
                            Editor.error("Err! Duplicate name:" + filename + "  form:path1:" + saveData[filename] + "  path2:" + resourcesPath);
                            return;
                        } else {
                            saveData[filename] = resourcesPath;
                        }
                        break;
                    }
                }

            }
        }
    },

    moveWXFile: function (uuidList) {

        self.moveDir(self.wxSrcDir, self.wxDstDir, uuidList);
    },

    uploadWXFile: function () {
        const OSS = require('ali-oss');
        const sPath = path.join(Editor.Project.path, "build", "wechatgame", "game.js");
        const ossPath = utils.getOSSBucketPath(sPath);
        const client = new OSS({
            region: Const.OSS_REGION,
            accessKeyId: Const.OSS_ACCESS_KEY,
            accessKeySecret: Const.OSS_ACCESS_KEY_SECRET,
            bucket: Const.OSS_BUCKET,
        });

        const srcDir = self.wxDstDir;
        const dstDir = path.posix.join(ossPath, "res");

        Editor.info('Put OSS [' + dstDir + "] from [" + srcDir + "]");
        let putTaskList = [];
        let fileCount = 0;
        utils.walkFiles(srcDir, function (fullPath) {
            let rPath = fullPath.substring(srcDir.length + 1).replace(/\\/g, "/"); // 这个地方要配合oss 格式，必须写/
            let dstFile = path.posix.join(dstDir, rPath);
            let task = () => {
                return client.put(dstFile, fullPath).then(function () {
                    fileCount++;
                    Editor.log("Upload success: " + fileCount);
                }).catch(function (e) {
                    Editor.error(e);
                    client.put(dstFile, fullPath).then(function () {
                        Editor.warn("Retry Upload success: " + fileCount);
                    });
                });
            };
            putTaskList.push(task);
        });
        Editor.info("PutTask Upload Task " + putTaskList.length);
        utils.promiseAll(putTaskList, 20, function () {
            Editor.info("Upload File finished " + fileCount);
        });
    },

    moveDir(srcPath, tarPath, filter = []) {
        fse.removeSync(tarPath);
        fse.ensureDirSync(tarPath);
        let dirs = fs.readdirSync(srcPath);
        dirs.forEach(function (filename) {
            let fileDir = path.join(srcPath, filename);
            let stats = fs.statSync(fileDir);
            let isFile = stats.isFile();
            if (isFile) { // 复制文件
                let filterFlag = filter.some(item => utils.startWith(filename, item));
                const destPath = path.join(tarPath, filename);
                fse.copySync(fileDir, destPath, {
                    overwrite: true
                });
                if (!filterFlag) {
                    fse.removeSync(fileDir);
                }
            } else { // 创建文件夹
                let tFileDir = path.join(tarPath, filename);
                fse.ensureDirSync(tFileDir);
                self.moveDir(fileDir, tFileDir, filter); // 递归
            }
        });

    },

};

module.exports = self;